/*
 *  compare.c - comparing functions
 *  Copyright (c)2002 - 2004 Andrej  OM1CW
 *  Copyright (c)2005 Mike Karas OM4AA <mike.om4aa@gmail.com>
 *  Copyright (c)2009 Ladislav Vaiz OK1ZIA <ok1zia@nagano.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "osdep.h"
#include "debug.h"
#include "main.h"
#include "call.h"
#include "compare.h"

/* used in make_band_refs() */
int compare_fcn2(const void *arg1, const void *arg2)
{
	DEBUG_FUNC("compare_fcn2");

	struct QSO_INFO **q1 = (struct QSO_INFO **) arg1;
	struct QSO_INFO **q2 = (struct QSO_INFO **) arg2;
	uint8_t q1ok = 1, q2ok = 1;

	if ((*q1)->r_call != (*q1)->ok_call)
		q1ok = 0;
	if ((*q2)->r_call != (*q2)->ok_call)
		q2ok = 0;

	if (q1ok && !q2ok)
		return -1;
	if (!q1ok && q2ok)
		return 1;

    //printf("compare %s=%d %s=%d\n", (*q1)->r_call->call, (*q1)->r_call->refs, (*q2)->r_call->call, (*q2)->r_call->refs);

	if ((*q1)->r_call->refs < (*q2)->r_call->refs)
		return -1;
	if ((*q1)->r_call->refs > (*q2)->r_call->refs)
		return 1;

	if (q1ok && q2ok)
		return (strcmp((*q1)->r_call->call, (*q2)->r_call->call));

	return (strcmp((*q1)->ok_call->call, (*q2)->ok_call->call));
}

int compare_fcn3(const void *arg1, const void *arg2)
{
	DEBUG_FUNC("compare_fcn3");

	struct CALL_INFO **c1 = (struct CALL_INFO **) arg1;
	struct CALL_INFO **c2 = (struct CALL_INFO **) arg2;

    //printf("compare3 %s=%d %s=%d\n", (*c1)->call, (*c1)->refs, (*c2)->call, (*c2)->refs);
	float e1 = (*c1)->refs ? ((float) (*c1)->temp * 100 / (*c1)->refs) : 0;
	float e2 = (*c2)->refs ? ((float) (*c2)->temp * 100 / (*c2)->refs) : 0;


	return ((/*uint32_t*/int) ((e2 - e1) * 1000));	/* TODO: need better type *
										 * conversion ? */
}

FILE *open_repfile(uint8_t band, int append) {
	static FILE *fd;
	float bnd = band2freq[band];
	char fname[BUFF_SIZE + 1];

	if (band >= BAND_GHZ && band < BAND_GHZ_ZERO)
		bnd *= 1000;

	if (band >= BAND_GHZ_ZERO)
		sprintf(fname, "%s%slx_G%d%s", cfg_epath, DIR_DELIMITER,
				(int) (bnd + 0.5), REPAIRS_EXT);
	else
		sprintf(fname, "%s%slx_%04d%s", cfg_epath, DIR_DELIMITER,
				(int) (bnd + 0.5), REPAIRS_EXT);

	if ((fd = fopen(fname, append ? "at+" : "wt")) == NULL)
		fd = stderr; /* if problem occured, use standard error output */

	return fd;
}

void close_repfile(FILE *fd) {
	if (fd != stderr && fd)
		fclose(fd);
}

inline void make_loc_cmp(uint8_t band)
{
	DEBUG_FUNC("make_loc_cmp");

	struct CALL_INFO *c = first_call[band], *c2, *last, *c3/*, *call_max_refs = NULL*/;
	struct QSO_INFO *q, *q2;
	uint8_t have_qso;
	char l[20];
	int found;
	FILE *fd;

	// printf("+ Same LOC check: ");

	printf(".");

	/* search call with same locs ??? */
	while (c) {
		if ((!c->first_call_same_loc) /* && c->refs */ &&c->loc) {
			c2 = c->next;
			c->first_call_same_loc = c;
			last = c;
			while (c2) {
				if ((!c2->first_call_same_loc)	/* && c2->refs */
					&&c->loc == c2->loc) { /* locs is the same */
					c2->first_call_same_loc = c; /* previous call with the same loc */
					last->next_call_same_loc = c2; /* next call with the same loc */
					last = c2;
				}
				c2 = c2->next;
			}
		}
		c = c->next;
	}

	c = first_call[band];
	
	fd = open_repfile(band, 1);
	fprintf(fd, "\n\nMaking comparision by same locators (calls without log):\n");

	while (c) {
		q = c->first_qso;
		while (q) {
			if (!q->ok_qso && q->ok_call->loc && /* q->ok_call->refs>0 && */
				q->ok_call->refs <= band2repairlimit[band]) {
//                printf("\n142: %s q->ok_call->refs=%d <= %d\n", q->ok_call->call, q->ok_call->refs, band2repairlimit[band]);
				c2 = q->ok_call->first_call_same_loc; /* trace calls with the same loc */
				while (c2) {
					if (c2 != q->ok_call && q->ok_call->loc
						&& q->ok_call->refs < c2->refs && c2->refs >= 3) { /* c2 possible correct call */
						if (compare_calls(q->ok_call->call, c2->call)) { /* calls equals is possible */
							q2 = c->first_qso; /* check if I don't have QSO with this station */
							have_qso = 0;
							if (q->ok_call->refs > 2) {
								while (q2) {
									if (q != q2 && c2 == q2->ok_call) {
										have_qso = 1; /* c2 already have qso with me */
										break;
									}
									q2 = q2->next;
								}
							}
							if (!have_qso) {
//                                printf("150\n");
								if (!c2->first_qso /*&& !q->r_call->first_qso*/ && verbose_mode > 1 && c->categ > 0) {
//                                    printf("\nq->ok_call->refs=%d\n", q->ok_call->refs);
									fprintf(fd, "%s: %s (%s,%d) ", c->call, q->ok_call->call, num2loc(q->ok_call->loc, l), q->ok_call->refs);
									fprintf(fd, "REPAIR: %s (%s,%d)", c2->call, num2loc(c2->loc, l), c2->refs);
									if (q->ok_call->refs > 1) {
										if (((int) (c2->refs / 2)) <= q->ok_call->refs)
											fprintf(fd, " [may be OK!]");
										else
											fprintf(fd, " [!]");
									}
									if (!compare_calls2(q->ok_call->call, c2->call))
										fprintf(fd, " [*]");
/*									if (q->ok_call->refs > 1 && !compare_calls2(q->ok_call->call, c2->call))
										fprintf(fd, " (may be OK!)");*/

									/* search for any other ok_call->call references */
									if (verbose_mode > 2) {
										c3 = first_call[band];
										found = 0;
										while (c3) {
											if (c3 != c2 && c3 != q->ok_call && compare_calls2(q->ok_call->call, c3->call)) {
												if (!found) {
													fprintf(fd, "\nFound OTHER referencies: ");
													found = 1;
												}
												fprintf(fd, "%s (%s,%d);", c3->call, num2loc(c3->loc, l), c3->refs);

												/* get max. references count */
												/*if ((!call_max_refs || call_max_refs->refs < c3->refs) && compare_locs(q->ok_call->loc, c3->loc))
													call_max_refs = c3;*/
											}
											c3 = c3->next;
										}

/*										c3 = call_max_refs;
										if (c3 && c3->refs > c2->refs)
											fprintf(fd, "FORCING REPAIR: %s (%s,%d)  !NOT IMPL - TODO!\n", c3->call, num2loc(c3->loc, l), c3->refs);
										else if (c3)
											fprintf(fd, "REQUESTED REPAIR OK\n");

										if (found) {
											if (!c3)
												fprintf(fd, "LOCs is too DIFFERENT, REQUESTED REPAIR OK\n");
											fprintf(fd, "-------------------------------------------------------------");
										}*/
									}
									fprintf(fd, "\n");
								}
								q->ok_call = c2;
                            //    printf("215 ok_call = %s\n", q->ok_call->call);
							}
						}
					}
					c2 = c2->next_call_same_loc;
				}
			}
			q = q->next;
		}
		c = c->next;
	}

	// printf("OK\n");
	close_repfile(fd);
}

inline void make_call_cmp(uint8_t band)
{
	DEBUG_FUNC("make_call_cmp");

	struct CALL_INFO *c = first_call[band], *c2, *last, *c3;
	struct QSO_INFO *q;
	char l[20];
	FILE *fd;
	int found;

	// printf("+ Same CALL check: ");

	printf(".");
	
	while (c) {
		if ((!c->first_call_same_call) /* && c->refs && c->loc */) {
			c2 = c->next;
			c->first_call_same_call = c;
			last = c;
			while (c2) {
				if ((!c2->first_call_same_call)	&& /* c2->refs && */
					compare_calls2(c->call, c2->call)) { /* cmp without '/' */
					c2->first_call_same_call = c;
					last->next_call_same_call = c2;
					last = c2;
				}
				c2 = c2->next;
			}
		}
		c = c->next;
	}

	c = first_call[band];
	
	fd = open_repfile(band, 0);
	fprintf(fd, "Making comparision by same calls (calls without log):\n");

	while (c) {
		q = c->first_qso;
		while (q) {
			if (!q->ok_qso /* && q->ok_call->loc && q->ok_call->refs>0 */
				&& q->ok_call->refs < 10) {
//                printf("\n266: %s q->ok_call->refs=%d < 10\n", q->ok_call->call, q->ok_call->refs);
				c2 = q->ok_call->first_call_same_call;
				while (c2) {
					if (c2 != q->ok_call /* && q->ok_call->loc */
						&& q->ok_call->refs < c2->refs) { /* c2 possible correct */
//                        printf("274: %d >= %d || %d\n", c2->refs, band2repairlimit[band], compare_locs(q->ok_call->loc, c2->loc));//fixme
						if (c2->refs >= band2repairlimit[band] || compare_locs(q->ok_call->loc, c2->loc)) {
							/* !!!! PRIDANE 8.11.02  , may be same locs */
							if (!c2->first_qso && verbose_mode > 1 && c->categ > 0) {
								fprintf(fd, "%s: %s (%s,%d) ", c->call, q->ok_call->call, num2loc(q->ok_call->loc, l), q->ok_call->refs);
								fprintf(fd, "REPAIR: %s (%s,%d)", c2->call, num2loc(c2->loc, l), c2->refs);
								if (q->ok_call->refs > 1) {
									if (((int) (c2->refs / 2)) <= q->ok_call->refs)
										fprintf(fd, " [may be OK!]");
									else
										fprintf(fd, " [!]");
								}
						
								if (verbose_mode > 2) {
									c3 = first_call[band];
									found = 0;
									while (c3) {
										if (c3 != q->ok_call && c3 != c2 && compare_locs(q->ok_call->loc, c3->loc) && compare_calls(q->ok_call->call, c3->call)) {
											if (!found) {
												fprintf(fd, "\nFound OTHER referencies: ");
												found = 1;
											}
											fprintf(fd, "%s (%s,%d);", c3->call, num2loc(c3->loc, l), c3->refs);
										}
										c3 = c3->next;
									}
								}
								fprintf(fd, "\n");
							}
							q->ok_call = c2;
//                            printf("308 ok_call = %s\n", q->ok_call->call);
						}
					}
					c2 = c2->next_call_same_call;
				}
			}
			q = q->next;
		}
		c = c->next;
	}

	// printf("OK\n");
	close_repfile(fd);
}

int compare_fcn(const void *arg1, const void *arg2)
{
	DEBUG_FUNC("compare_fcn");

	struct CALL_INFO **c1 = (struct CALL_INFO **) arg1;
	struct CALL_INFO **c2 = (struct CALL_INFO **) arg2;

	return ((*c2)->score - (*c1)->score);
}

int compare_claimed_fcn(const void *arg1, const void *arg2)
{
	DEBUG_FUNC("compare_claimed_fcn");

	struct CALL_INFO **c1 = (struct CALL_INFO **) arg1;
	struct CALL_INFO **c2 = (struct CALL_INFO **) arg2;

	return ((*c2)->score_claimed - (*c1)->score_claimed);
}

/* returns 1 if calls is equal with few differenced chars */
uint8_t compare_calls(char *call1, char *call2)
{
	DEBUG_FUNC("compare_calls");

	int len1 = strlen(call1);
	int len2 = strlen(call2);
	int len, i;
	int err = 0;

	/* len = min(len1, len2); */
	len = len1;
	if (len > len2)
		len = len2;

    /* in how much chars calls differs? */
	for (i = 0; i < len; i++)
		if (call1[i] != call2[i])
			err++;

//    printf("  compare_calls('%s','%s') len1=%d  len2=%d  err=%d\n", call1, call2, len1, len2, err) ;

	/* same lengths */
	if (len1 == len2 && len > 4 && err < 3)
		return 1;
	if (len1 == len2 && len > 3 && err < 2)
		return 1;
	/* lengths differs */
	if (abs(len1 - len2) <= 2 && !err)
		return 1;
	if (abs(len1 - len2) < 2 && err < 2)
		return 1;

	return 0; /* too much differences */
}

/* returns 1 if locs differs in max 2 chars */
uint8_t compare_locs(uint32_t loc1, uint32_t loc2)
{
	DEBUG_FUNC("compare_locs");

	char l1[7], l2[7];
	int err = 0, i;

	if (!loc1 || !loc2)
		return 0;
	if (loc1 == loc2)
		return 1;

	num2loc(loc1, l1);
	num2loc(loc2, l2);

	for (i = 0; i < 6; i++)
		if (l1[i] != l2[i])
			err++;

//    printf("  compare_locs('%s','%s')  err=%d\n", loc1, loc2, err) ;
    
	if (err < 3)
		return 1;

	return 0;
}

/* compare calls without any '/' (strong) after call */
uint8_t compare_calls2(char *call1, char *call2)
{
	DEBUG_FUNC("compare_calls2");

	char call11[SIZE_CALL + 1];
	char call22[SIZE_CALL + 1];
	char *c;
    int ret = 0;

	if (!strcmp(call1, call2)){
        ret = 1;
        goto x;
    }

	strcpy(call11, call1);
	c = strchr(call11, 0);
	if ((c - 3) < call11){
		goto x; /* too short call */
    }
	/* remove '/' */
	if (*(--c) == '/')
		*c = 0;
	else if (*(--c) == '/')
		*c = 0;
	else if (*(--c) == '/')
		*c = 0;

	strcpy(call22, call2);
	c = strchr(call22, 0);
	if ((c - 3) < call22){
		goto x; /* too short */
    }
	/* remove '/' */
	if (*(--c) == '/')
		*c = 0;
	else if (*(--c) == '/')
		*c = 0;
	else if (*(--c) == '/')
		*c = 0;

    
	/* compare again */
	if (!strcmp(call11, call22)){
        ret = 1;
        goto x;
    }
    
    //printf("  compare_calls2('%s','%s')  \n", call1, call2) ;
x:;
	return ret;
}


static int min3(int a, int b, int c)
{
    int min = a;
    if (b < min) min=b;
    if (c < min) min=c;
    return min;
}

// Levensthein distance between strings
int levenshtein(const char *s, const char *t){
    int i, j, k, n, m, cost, *d, ret;
    n = strlen(s); 
    m = strlen(t);
    if (!n) return -1;
    if (!m) return -1;
    
    // d is a table with n+1 rows and m+1 columns
    d = malloc(sizeof(int) * (m + 1) * (n + 1));
    m++;
    n++;
    // step 2  
    for(k = 0; k < n; k++) d[k] = k;    // deletion
    for(k = 0; k < m; k++) d[k*n] = k;  // insertion
    // step 3 and 4    
    for(i = 1; i<n; i++)
    for(j=1; j<m; j++){
        // step 5
        if(s[i - 1] == t[j - 1])
            cost = 0;
        else
            cost = 1;
        // step 6             
        d[j * n + i] = min3(
            d[(j - 1) * n + i] + 1,
            d[j * n + i - 1] + 1,
            d[(j - 1) * n + i - 1] + cost);
    }
    ret = d[n * m - 1];
    free(d);
    return ret;
}


